unit HallSel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, HotImage, StdCtrls;

type
  TfrmHallSel = class(TForm)
    ScrollBox1: TScrollBox;
    HotImage1: THotImage;
    Panel1: TPanel;
    Label1: TLabel;
    Button1: TButton;
    Button2: TButton;

    procedure HotImage1GetHintText(Sender: TObject; Spot: TSpot;
      var Text: String);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure HotImage1Paint(Sender: TObject);
    procedure HotImage1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ScrollBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
    FSpots:TList; //list of currently selected spots
    FRgn:HRgn;//Windows "region" for painting current selection
    procedure LoadData;
    procedure DrawSelection(rgn:HRgn);
    procedure SetSelection(ASpot:TSpot);
    procedure AddSelection(ASPot:TSpot);
    procedure CancelSelection;
  public
    { Public declarations }
  end;

var
  frmHallSel: TfrmHallSel;

implementation

uses sellist;

{$R *.DFM}

{procedure for drawing selection }
procedure TfrmHallSel.DrawSelection(Rgn:HRgn);
begin
//note - here is painting directly to the screen, not bitmap
 With HotImage1.Canvas do
  if Rgn<>0 then
  begin
   Pen.Mode:=pmNotXor; //inverted drawing
    FrameRgn(Handle,Rgn,GetStockObject(LTGRAY_BRUSH),4,4);
  end;
end;

procedure TfrmHallSel.SetSelection(ASpot:TSpot);
begin
 DrawSelection(FRgn);
 DeleteObject(FRgn);
 FSpots.Clear;
 FRgn:=HotImage1.CreateRgnHandle(ASpot);
 FSpots.Add(ASpot);
 DrawSelection(FRgn);
end;

procedure TfrmHallSel.AddSelection(ASPot:TSpot);
var
 tempRgn:HRgn;
 i:integer;
const
 combineMap:array[boolean]of integer=(RGN_OR,RGN_DIFF);

begin
 if FSpots.Count=0 then SetSelection(ASpot) else

 With HotImage1 do
 begin
  tempRgn:=HotImage1.CreateRgnHandle(ASpot);
  i:=FSpots.IndexOf(ASpot);
  CombineRgn(FRgn,FRgn,tempRgn,CombineMap[i>=0]);
  DrawSelection(tempRgn);
  DeleteObject(tempRgn);
  if i>=0 then FSpots.Delete(i)
  else FSpots.Add(ASpot);
 end;
end;

procedure TfrmHallSel.CancelSelection;
begin
 DrawSelection(FRgn);
 if FRgn<>0 then DeleteObject(FRgn);
 FRgn:=0;
 FSpots.Clear; //remove all selection
end;


procedure TfrmHallSel.LoadData;
var
 PictHeight,PictWidth:integer;

procedure AddSite(X,Y:integer;const SiteNo:string;
                   SiteID:integer;BrushColor:TColor;Hot:boolean);
var R:TRect;
    Spot:TSpot;
begin
 R:=Rect(X,Y,X+40,y+40);
 {To draw on picture(not to screen directly)-
       use PictureCanvas instead of Canvas}
 With HotImage1.PictureCanvas do
 begin
  Brush.Color:=BrushColor;
  FillRect(R);
  DrawText(Handle,PChar(SiteNo),
           length(SiteNo),R,DT_CENTER or DT_VCENTER or DT_SINGLELINE);
  if Hot then
  begin
   Spot:=TSpot.Create(skRectangle);
   Spot.SpotID:=SiteID;
   Spot.HintText:=SiteNo;
   Spot.BoundsRect:=R;
   HotImage1.HotSpots.AddSpot(Spot);
  end;
 end;
end;
{better to know total picture rectangle BEFORE loading data - it's vy time-consuming
  operation to do that calculation each time after all data loaded.
  Store that size in database.
   if you have a problem with that - write me }
begin
 PictHeight:=400;
 PictWidth:=500;


 With HotImage1 do
 begin
  CancelSelection;
  Picture.Graphic:=nil; //invalidate all picture
  BoundsRect:=Bounds(0,0,PictWidth,PictHeight);//Set Size of HotImage
  PictureCanvas.Brush.Color:=clBlack;
  PictureCanvas.FillRect(BoundsRect);
  {you can load your background picture instead}

 end;
{ assume loading from database}

  AddSite(10,10,'1',1{perhaps Row num},clWhite,true);
  AddSite(60,10,'2',1,clWhite,true);
  AddSite(110,10,'3',1,clWhite,true);
  AddSite(160,10,'4',1,clRed,false);
  AddSite(210,10,'5',1,clWhite,true);
  AddSite(260,10,'6',1,clWhite,true);
  AddSite(260,10,'7',1,clWhite,true);
  AddSite(310,10,'8',1,clWhite,true);
  AddSite(360,10,'9',1,clWhite,true);

  AddSite(25,60,'1',2{perhaps Row num},clWhite,true);
  AddSite(75,60,'2',2,clWhite,true);
  AddSite(125,60,'3',2,clWhite,true);
  AddSite(175,60,'4',2,clGreen,false);
  AddSite(225,60,'5',2,clGreen,false);
  AddSite(275,60,'6',2,clWhite,true);
  AddSite(325,60,'7',2,clWhite,true);



end;




procedure TfrmHallSel.HotImage1GetHintText(Sender: TObject; Spot: TSpot;
  var Text: String);
begin
  Text:=Format('Raw #%d, site "%s"',[Spot.SPotID,Spot.HintText]);
end;

procedure TfrmHallSel.FormCreate(Sender: TObject);
begin
 FSpots:=TList.Create; //list of pointers
 LoadData;
end;

procedure TfrmHallSel.FormDestroy(Sender: TObject);
begin
 DeleteObject(FRgn);
 FSpots.Free;
end;

procedure TfrmHallSel.HotImage1Paint(Sender: TObject);
begin
 DrawSelection(FRgn);
end;

procedure TfrmHallSel.HotImage1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin

 with HotImage1 do
  if ssCtrl in Shift then AddSelection(HotSpots[SpotIndex])
  else SetSelection(HotSpots[SpotIndex]);


end;

procedure TfrmHallSel.ScrollBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 if not (ssCtrl in Shift) then
                 CancelSelection;
end;

procedure TfrmHallSel.Button1Click(Sender: TObject);
var i:integer;
begin
 with TfrmSelList.Create(Application)do
 try
  For i:=0 to FSpots.Count-1 do
  With TSpot(FSpots[i])do
   ListBox1.Items.Add(Format('raw=%d,Num=%s',[SpotID,HintText]));
  ShowModal;
 finally
  free;
 end;
end;

end.
