unit HotImage;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

{$R-}
type
  PPoints = ^TPoints;
  TPoints = array[0..1] of TPoint;
  TSpotKind=(skPolyline,skRectangle,skElipse);

  THotSpots = class;
  THotImage = class;



  TSpot = class
  private
   FPoints:TList;
   FPolyRgn:HRgn;
   FSpotKind:TSpotKind;
   FHintText:string;
   FID:integer;
   FEnabled:boolean;
   function GetRegion:HRgn;
   procedure ReleaseRegion;
   function GetPoint(Index:integer):Tpoint;
   procedure SetPoint(Index:integer;const APoint:TPoint);
   procedure SetSpotKind(Value:TSpotKind);

   //bounds manupulation
   function GetBoundsRect:TRect;
   procedure SetBoundsRect(const R:TRect);
 
  public

   //Constructors, destructors
   constructor Create(ASpotKind:TSpotKind);
   destructor Destroy;override;


   procedure Assign(Source:TSpot);

   //returns pointer to array[0..n] of TPoint
   function GetPointArray:PPoints;

   //IS point in spot?
   function PointInSpot(const APoint:TPoint):boolean;

   //number of points in spot
   function Count:integer;
   
   
   
   //Clear all points
   procedure Clear;

   //Bounds manipulation
 
  procedure Offset(dx,dy:integer);


   //Point manipulation
   procedure AddPoint(const APoint:TPoint);
   procedure DeletePoint(Index:Integer);
   procedure InsertPoint(Index:integer;const APoint:TPoint);
   property Points[Index:integer]:TPoint read GetPoint Write SetPoint; default;

   property RgnHandle:HRgn read GetRegion;


   property HintText:string read FHintText write FHintText;
   property SpotID:integer read FID write FID;
   property Enabled:boolean read FEnabled write FEnabled;
   property SpotKind:TSpotKind read FSpotKind write SetSpotKind;
   property BoundsRect:TRect read GetBoundsRect write SetBoundsRect;
  end;


  THotSpots=class(TPersistent)
  private
   FList:TList;
   FOwner:THotImage;
   function GetSpot(Index:integer):TSpot;
   procedure SetSpot(index:integer;ASpot:TSpot);
   procedure WriteSpots(Stream:TStream);
   procedure ReadSpots(Stream:TStream);
  protected
   procedure DefineProperties(Filer: TFiler);override;
   procedure Assign(Source:TPersistent);override;
  public
   constructor Create(AOwner:THotImage);
   destructor Destroy;override;

   function GetBoundsRect:TRect;

   //Streaming routines
   procedure SaveToStream(Stream:TStream);
   procedure LoadFromStream(Stream:TStream);

   //Spot manipulations;
   procedure DeleteSpot(Index:integer);
   function AddSpot(ASpot:TSpot):integer;
   function RemoveSpot(Index:integer):TSpot;
   procedure InsertSpot(Index:integer;ASpot:TSpot);
   procedure MoveSpot(CurrentIndex,NewIndex:integer);
   procedure Clear;

   function GetSpotByID(ASpotID:integer):TSpot;

   //Coordinates routines
   function IndexByPoint(const Point:TPoint):integer;
   function SpotByPoint(const Point:TPoint):TSpot;


   function Count:integer;

   property Spots[Index:integer]:TSpot read GetSpot write SetSpot; default;
   property Owner:THotImage read FOwner;

  end;
  TGetHintTextEvent=procedure (Sender:TObject;Spot:TSpot;var Text:String)of object;
  THotImage = class(TImage)
  private
    { Private declarations }
   FHotSpots:THotSpots;
   FSpotIndex:integer;
   FCommonHint:string;
   FOnPaint:TNotifyEvent;
   FOnGetHintText:TGetHintTextEvent;
   procedure DoPaint;
   function GetCanvas:TCanvas;
   function GetPictureCanvas:TCanvas;
   {$IFNDEF VER100}
   function DestRect:TRect;
 {$ENDIF}
  protected
    { Protected declarations }
    procedure CMHintShow(var Message:TMessage);message CM_HINTSHOW;
    procedure CMHitTest(var Message:TCMHitTest);message CM_HITTEST;
    procedure Paint;override;
  public
    { Public declarations }
    constructor Create(AOwner:TComponent);override;
    destructor Destroy;override;

    function CopyPointArray(PointBuffer:PPoints;Spot:TSpot):boolean;
    function CreateRgnHandle(Spot:TSpot):HRgn;
    function CreateRgnByID(ASpotID:integer):HRgn;

    procedure SetHotSpots(AHotSpots:THotSpots);
    property SpotIndex:integer read FSpotIndex;
    property Canvas:TCanvas read GetCanvas;
    property PictureCanvas:TCanvas read GetPictureCanvas;
  published
    { Published declarations }
    property HotSpots:THotSpots read FHotSpots write SetHotSpots;
    property Hint read FCommonHint write FCommonHint;     //hides original TControl.Hint

    {$IFDEF VER100}
    property Cursor default crHandPoint;
    {$ENDIF}

    {Events}
    property OnPaint:TNotifyEvent read FOnPaint write FOnPaint;
    property OnGetHintText:TGetHintTextEvent read FOnGetHintText write FOnGetHintText;
  end;


implementation


{*** TSpot ***}


{modified 19.02.98}
constructor TSpot.Create(ASpotKind:TSpotKind);
begin
 FPoints:=TList.Create;
 FEnabled:=true;
 FSpotKind:=ASpotKind;
 {all SpotKinds except skPolyline have fixed number of points}
 if ASpotKind<>skPolyLine then
 FPoints.Count:=4;

end;

destructor TSpot.Destroy;
begin
 ReleaseRegion;
 FPoints.Free;
end;

procedure TSpot.ReleaseRegion;
begin
 If FPolyRgn<>0 then DeleteObject(FPolyRgn);
 FPolyRgn:=0;
end;

function TSpot.GetRegion:HRgn;
var tempRect:PRect;
begin
 if FPolyRgn=0 then
 begin
 if FSpotKind= skPolyline then
  FPolyRgn:=CreatePolygonRgn(GetPointArray^,Count,WINDING)
  else
  begin
   TempRect:=PRect(GetPointArray);
   with TempRect^ do
   case FSpotKind of
    skRectangle:FPolyRgn:=CreateRectRgn(Left,Top,Right,Bottom);
    skElipse:FPolyRgn:=CreateEllipticRgn(Left,Top,Right,Bottom);
   end;
  end;
 end;
 result:=FPolyRgn;
end;

function TSpot.Count:integer;
begin
 result:=FPoints.Count div 2;
end;

procedure TSpot.Clear;
begin
  ReleaseRegion;
  SetSpotKind(skPolyLine);
  FPoints.Clear;
end;

procedure TSpot.InsertPoint(Index:integer;const APoint:TPoint);
begin
  Index:=index*2;
   ReleaseRegion;

  SetSpotKind(skPolyLine);

  FPoints.Insert(index,pointer(APoint.Y));
  FPoints.Insert(index,pointer(APoint.X));
end;

function TSpot.GetPoint(Index:integer):TPoint;
begin
 Index:=index*2;
 With result,FPoints do
 begin
  X:=longint(Items[Index]);
  Y:=longint(Items[Index+1]);
 end;
end;

{added 19.02.98}
procedure TSpot.SetSpotKind(Value:TSpotKind);
begin
 if value<>FSpotKind then
 begin
  ReleaseRegion;
  if Value<>skPolyline then FPoints.Count:=4;
  FSpotKind:=Value;
 end;
end;



procedure TSpot.SetPoint(Index:integer;const APoint:TPoint);
begin
  Index:=Index*2;
  ReleaseRegion;
  FPoints.Items[index]:=pointer(APoint.X);
  FPoints.Items[index+1]:=pointer(APoint.Y);
end;

procedure TSpot.AddPoint(const APoint:TPoint);
begin
  ReleaseRegion;
  SetSpotKind(skPolyLine);
  FPoints.Add(Pointer(APoint.X));
  FPoints.Add(Pointer(APoint.Y));
end;





procedure TSpot.DeletePoint(Index:Integer);
begin
  Index:=Index*2;
  ReleaseRegion;
  SetSpotKind(skPolyLine);
  FPoints.Delete(Index);
  FPoints.Delete(Index);
end;

function TSpot.GetPointArray:PPoints;
begin
 result:=PPoints(@FPoints.List[0]);
end;

function TSpot.PointInSpot(const APoint:TPoint):boolean;
begin
{if spotkind have rect area - simple call PtInRect(without creating regions)}
if FSpotkind=skRectangle then
 result:=PtInRect(PRect(GetPointArray)^,APoint)
else with APoint do
      result:=PtInRegion(GetRegion,X,Y);


 {if not result then
  ReleaseRegion;}
end;


{bounds manupulation}
function TSpot.GetBoundsRect:TRect;
begin
 if FSpotKind=skPolyline then
 begin
  if GetRgnBox(GetRegion,result)=0 then fillchar(result,sizeof(result),0);

 end
 else result:=PRect(GetPointArray)^;
end;

procedure TSpot.SetBoundsRect(const R:TRect);
var 
 OldRect:TRect;
 ptCount:integer;
 P:PPoint;
 Offs,Nom,Denom:TPoint;
begin
 if FSpotKind=skPolyline then 
 begin
  OldRect:=GetBoundsRect;
  ReleaseRegion;

  Offs.X:=R.Left-OldRect.Left;
  Offs.Y:=R.Top-OldRect.Top;
  Nom.X:=R.Right-R.Left;
  Nom.Y:=R.Bottom-R.Top;
  Denom.X:=OldRect.Right-OldRect.Left;
  Denom.Y:=OldRect.Bottom-OldRect.Top;


  ptCount:=Count;
  p:=PPoint(GetPointArray);
  while ptCount>0 do
  begin
   P^.x:=Offs.x+Muldiv(p^.x,Nom.x,Denom.x);
   P^.y:=Offs.y+Muldiv(p^.y,Nom.y,Denom.y);
   inc(p);
   dec(ptCount);
  end;
 end
 else
 begin
  ReleaseRegion;
  PRect(GetPointArray)^:=R;
 end;
end; 

procedure TSpot.Offset(dx,dy:integer);
var p:PPoint;
    ptCount:integer;
begin
 ReleaseRegion;
 ptCount:=Count;
 p:=PPoint(GetPointArray);
 while ptCount>0 do
 begin
  with p^ do
   begin
    inc(x,dx);
    inc(y,dy);
   end;
  inc(p);
  dec(ptCount);
 end;
end;

procedure TSpot.Assign(Source:TSpot);
var
   cnt:integer;
begin
 if Source<>nil then
 begin
   Clear;
   With Source.FPoints do
   begin
    cnt:=Count;
    Self.FPoints.Count:=cnt;
    system.Move(List[0],Self.FPoints.List[0],cnt*sizeof(pointer));
   end;
  SetSpotKind(Source.FSpotKind);
  HintText:=Source.HintText;
  Enabled:=Source.Enabled;
  SpotID:=Source.SpotID;
 end;
end;


{*** THotSpots ***}
constructor THotSpots.Create(AOwner:THotImage);
begin
 inherited Create;
 FOwner:=AOwner;
 FList:=TList.Create;

end;

destructor THotSpots.Destroy;
begin
 Clear;
 FList.Free;
 inherited Destroy;
end;

procedure THotSpots.Clear;
var
 i:integer;
begin
 For i:=0 to FList.Count-1 do
  TSpot(FList.Items[i]).Free;
 FList.Clear;
end;


function THotSpots.GetSpot(Index:integer):TSpot;
begin
 result:=TSpot(FList.Items[Index]);
end;

procedure THotSpots.SetSpot(index:integer;ASpot:TSpot);
begin
 TSpot(FList.Items[Index]).assign(ASpot);
end;

{function PointToString(const Point:TPoint):string ;
begin
 result:=format('%d,%d',[Point.X,Point.Y]);
end;


function StringToPoint(const S:string):TPoint;
var DelimPos:integer;
begin
 DelimPos:=Pos(',',S);
 With result do
 begin
   X:=StrToIntDef(Copy(S,1,DelimPos-1),0);
   Y:=StrToIntDef(Copy(S,DelimPos+1,Length(s)),0);
 end;
end;
 }

procedure THotSpots.WriteSpots(Stream:TStream);
var Spot:TSpot;

procedure WritePoints;
var
    N:integer;
begin
 With Stream do
 begin
  N:=Spot.Count;
  WriteBuffer(N,Sizeof(N));
  WriteBuffer(Spot.GetPointArray^,sizeof(TPoint)*N);
 end;
end;

procedure WriteRect;
begin
 Stream.WriteBuffer(Spot.GetPointArray^,Sizeof(TRect));
end;

var
    i,N,tmp:integer;
    SK:TSpotKind;
begin
  With Stream do
  begin
   N:=0;//indicating that new stream version loaded
   WriteBuffer(N,Sizeof(N));
   N:=FList.Count;
   WriteBuffer(N,Sizeof(N));
   for i:=0 to N-1 do
   begin
    Spot:=Spots[i];
    SK:=Spot.FSpotKind;
    Writebuffer(SK,1); //only one byte needed
    WriteBuffer(Spot.FID,sizeof(integer));
    WriteBuffer(Spot.FEnabled,sizeof(boolean));
    tmp:=length(Spot.FHintText);
    WriteBuffer(tmp,sizeof(integer));
    if tmp<>0 then
    WriteBuffer(PChar(Spot.FHintText)^,tmp);
    if SK=skPolyline then
     WritePoints
    else WriteRect

   end;
  end;
end;

procedure THotSpots.ReadSpots(Stream:TStream);
var
 Spot:TSpot;
 NewVersion:boolean;
procedure ReadPoints;
var
   n:integer;

begin
 With Stream do
 begin
  ReadBuffer(N,sizeof(N));
  Spot.FPoints.Count:=n*2;
  ReadBuffer(Spot.GetPointArray^,sizeof(TPoint)*n);
 end;
end;

procedure ReadRect;
begin
  Stream.readBuffer(Spot.GetPointArray^,sizeof(TRect));
end;

var
 i,n,tmp:integer;
 SK:TSpotKind;
begin
Clear;
 With Stream do
 begin
  ReadBuffer(n,sizeof(N));
  NewVersion:=n=0;
  if NewVersion then
   ReadBuffer(n,sizeof(N));
  For i:=0 to n-1 do
  begin
   SK:=skPolyline;
   if NewVersion then ReadBuffer(SK,1);
   Spot:=TSpot.Create(SK);

   ReadBuffer(Spot.FID,sizeof(integer));
   ReadBuffer(Spot.FEnabled,Sizeof(boolean));
   ReadBuffer(tmp,sizeof(integer));
   SetLength(Spot.FHintText,tmp);
   if tmp<>0 then
    ReadBuffer(PChar(Spot.FHintText)^,tmp);

   if sK=skPolyline then
    ReadPoints else readRect;
   AddSpot(Spot);
  end;
 end;
end;



procedure THotSpots.DefineProperties(Filer: TFiler);
begin
 inherited DefineProperties(Filer);
 Filer.DefineBinaryProperty('Spots',ReadSpots,WriteSpots,FList.Count>0);
end;


procedure THotSpots.Assign(Source:TPersistent);
var
  i:integer;
  ASpot:TSpot;
begin
  If (Source<>nil) and (Source is THotSpots) then
  begin
    Clear;
    For i:=0 to THotSpots(Source).FList.Count-1 do
    begin
      ASpot:=TSpot.Create(skPolyLine);
      ASpot.Assign(THotSpots(Source).Spots[i]);
      AddSpot(ASpot);
    end;
  end
   else inherited Assign(Source);
end;

procedure THotSpots.DeleteSpot(Index:integer);
begin
 TSpot(FList.Items[Index]).Free;
 FList.Delete(Index);
end;


function THotSpots.AddSpot(ASpot:TSpot):integer;
begin
 result:=FList.Add(ASpot);
end;

function THotSpots.RemoveSpot(Index:integer):TSpot;
begin
 result:=TSpot(FList.Items[Index]);
 FList.Delete(Index);
end;

procedure THotSpots.MoveSpot(CurrentIndex,NewIndex:integer);
begin
  FList.Move(CurrentIndex,NewIndex);
end;

procedure THotSpots.InsertSpot(Index:integer;ASpot:TSpot);
begin
 FList.Insert(Index,ASpot);
end;

function THotSpots.IndexByPoint(const Point:TPoint):integer;
var i,cnt:integer;

begin
 result:=-1;
 cnt:=FList.Count-1;
 For i:=0 to cnt do
 With Spots[i]do
  IF Enabled and PointInSpot(Point) then
  begin
   result:=i;
   break;
  end;
end;

function THotSpots.GetSpotByID(ASpotID:integer):TSpot;
var i:integer;
    Spot:TSpot;
begin
 result:=nil;
 for i:=0 to FList.Count-1 do
 begin
  Spot:=TSpot(FList.Items[i]);
  if Spot.FID =ASpotID then
  begin
   result:=Spot;
   break;
  end;
 end;
end;


function THotSpots.SpotByPoint(const Point:TPoint):TSpot;
var i:integer;
begin
  i:=IndexByPoint(Point);
  if i>=0 then result:=Spots[i] else
   result:=nil;
end;

function THotSpots.Count:integer;
begin
  result:=FList.Count;
end;




{streaming functions}
const HS_Sign_old:array[0..3]of char='HMAP';
      HS_Sign:array[0..3]of char='HMAp';

procedure THotSpots.SaveToStream(Stream:TStream);
procedure WriteSign;
begin
 Stream.WriteBuffer(HS_Sign,4);
end;
begin
 WriteSign;
 WriteSpots(Stream);
end;



procedure THotSpots.LoadFromStream(Stream:TStream);
procedure ReadSign;
var i:integer;
begin
 Stream.ReadBuffer(i,4);
 if (longint(HS_Sign)<>i) and (longint(HS_Sign_old)<>i)then
  raise EStreamError.Create('HotMap: Invalid Stream format');
end;

begin
  ReadSign;
  ReadSpots(Stream);
end;


function THotSpots.GetBoundsRect:TRect;
var
 P:PPoint;
 Spot:TSpot;
 PointCount:integer;
 i,j:integer;

procedure AdjustRect;
begin
with result,P^ do
begin
 if Left>X then Left:=X;
 if Right<X then Right:=X;
 if Top>Y then Top:=Y;
 if Bottom<Y then Bottom:=Y;
end;
end;

begin
FillChar(result,sizeof(TRect),0);
 if FList.Count>0 then
 begin
 i:=0;
  Repeat
   Spot:=Spots[i];
   PointCount:=Spot.Count;
   inc(i);
  until (PointCount<>0) or (i>=FList.Count);
  if i<FList.Count then exit;
  P:=PPoint(Spot.GetPointArray);
  Result.TopLeft:=P^;
  Result.BottomRight:=P^;
  For j:=1 to PointCount-1 do
  begin
   inc(P);
   AdjustRect;
  end;
  while i<FList.Count do
  begin
   Spot:=Spots[i];
   PointCount:=Spot.Count;
   P:=PPoint(Spot.GetPointArray);
   for j:=0 to PointCount-1 do
   begin
    AdjustRect;
    inc(P);
   end;
   inc(i);
  end;
 end;


end;


{*** THotImage ***}
constructor THotImage.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 FHotSpots:=THotSpots.Create(self);
 FSpotIndex:=-1;
{$IFDEF VER100}
 {"hand" cursor was defined since Delphi 3's VCL}
 Cursor:=crHandPoint;
{$ENDIF}
end;

destructor THotImage.Destroy;
begin
 FHotSpots.Free;
 inherited Destroy;
end;

procedure THotImage.SetHotSpots(AHotSpots:THotSpots);
begin
 FHotSpots.Assign(AHotSpots);
end;

function THotImage.CreateRgnHandle(Spot:TSpot):HRgn;
var
 Buffer:PPoints;
 BufferSize:Integer;
 sk:TSpotKind;
 tempRect:PRect;
begin

 result:=0;
 if Spot=nil then exit;
 BufferSize:=Spot.Count*sizeof(TPoint);
 if BufferSize>0 then
 begin
 sk:=Spot.FSpotKind;
 GetMem(Buffer,BufferSize);
 try
  if CopyPointArray(Buffer,Spot) then
  begin
  if sk=skPolyline then
    result:=CreatePolygonRgn(Buffer^,Spot.Count,WINDING)
  else
  begin
   SK:=Spot.FSpotKind;
   TempRect:=PRect(Buffer);
   with TempRect^ do
   case SK of
    skRectangle:result:=CreateRectRgn(Left,Top,Right,Bottom);
    skElipse:result:=CreateEllipticRgn(Left,Top,Right,Bottom);
   end;
  end;
 end;
 finally
  FreeMem(Buffer,BufferSize);
 end;
 end;
end;

function THotImage.CreateRgnByID(ASpotID:integer):HRgn;
var i:integer;
    Spot:TSpot;
    TmpRgn:HRgn;
begin
 result:=0;
 For  i:=0 to FHotSpots.Count-1 do
 begin
   Spot:=FHotSpots[i];
   if Spot.SpotID=ASpotID then
   begin
    TmpRgn:=CreateRgnHandle(Spot);
    if result=0 then result:=TmpRgn
    else
    begin
     CombineRgn(result,result,TmpRgn,RGN_OR);
     DeleteObject(TmpRgn);
    end;
   end;
 end;
end;

{$IFNDEF VER100}
function THotImage.DestRect:TRect;
begin
 if Stretch then
    Result := ClientRect
  else if Center then
    Result := Bounds((Width - Picture.Width) div 2, (Height - Picture.Height) div 2,
      Picture.Width, Picture.Height)
  else
    Result := Rect(0, 0, Picture.Width, Picture.Height);
end;

{$ENDIF}

function THotImage.CopyPointArray(PointBuffer:PPoints;Spot:TSpot):boolean;
var
 i:integer;
 N,D,Offset:TPoint;
 SrcPoints:PPoints;
 PointsCount:integer;
begin

 if (Spot=nil) or(Picture.Graphic=nil) then
 begin
  result:=false;
  exit;
 end;

 With DestRect do
 begin
  Offset.x:=Left;
  N.X:=Right-Left;
  Offset.y:=Top;
  N.Y:=Bottom-Top;
 end;
 With Picture.Graphic do
 begin
  D.X:=Width;
  D.Y:=Height;
 end;
  PointsCount:=Spot.Count-1;
  SrcPoints:=Spot.GetPointArray;
  for i:=0 to PointsCount do
   with SrcPoints[i] do
   begin
    PointBuffer[i].X:=MulDiv(X,N.X,D.X)+Offset.X;
    PointBuffer[i].Y:=MulDiv(Y,N.Y,D.Y)+Offset.Y;
   end;
 result:=true;
end;



procedure THotImage.CMHintShow(var Message:TMessage);


{$IFNDEF VER100}
{GetCursorHeightMargin from VCL in D3}
  { Return number of scanlines between the scanline containing cursor hotspot
    and the last scanline included in the cursor mask. }
  function GetCursorHeightMargin: Integer;
  var
    IconInfo: TIconInfo;
    BitmapInfoSize: Integer;
    BitmapBitsSize: Integer;
    Bitmap: PBitmapInfoHeader;
    Bits: Pointer;
    BytesPerScanline, ImageSize: Integer;

      function FindScanline(Source: Pointer; MaxLen: Cardinal;
        Value: Cardinal): Cardinal; assembler;
      asm
              PUSH    ECX
              MOV     ECX,EDX
              MOV     EDX,EDI
              MOV     EDI,EAX
              POP     EAX
              REPE    SCASB
              MOV     EAX,ECX
              MOV     EDI,EDX
      end;

  begin
    { Default value is entire icon height }
    Result := GetSystemMetrics(SM_CYCURSOR);
    if GetIconInfo(GetCursor, IconInfo) then
    try
      GetDIBSizes(IconInfo.hbmMask, BitmapInfoSize, BitmapBitsSize);
      Bitmap := AllocMem(BitmapInfoSize + BitmapBitsSize);
      try
        Bits := Pointer(Longint(Bitmap) + BitmapInfoSize);
        if GetDIB(IconInfo.hbmMask, 0, Bitmap^, Bits^) and
          (Bitmap^.biBitCount = 1) then
        begin
          { Point Bits to the end of this bottom-up bitmap }
          with Bitmap^ do
          begin
            BytesPerScanline := ((biWidth * biBitCount + 31) and not 31) div 8;
            ImageSize := biWidth * BytesPerScanline;
            Bits := Pointer(Integer(Bits) + BitmapBitsSize - ImageSize);
            { Use the width to determine the height since another mask bitmap
              may immediately follow }
            Result := FindScanline(Bits, ImageSize, $FF);
            { In case the and mask is blank, look for an empty scanline in the
              xor mask. }
            if (Result = 0) and (biHeight >= 2 * biWidth) then
              Result := FindScanline(Pointer(Integer(Bits) - ImageSize),
                ImageSize, $00);
            Result := Result div BytesPerScanline;
          end;
          Dec(Result, IconInfo.yHotSpot);
        end;
      finally
        FreeMem(Bitmap, BitmapInfoSize + BitmapBitsSize);
      end;
    finally
      if IconInfo.hbmColor <> 0 then DeleteObject(IconInfo.hbmColor);
      if IconInfo.hbmMask <> 0 then DeleteObject(IconInfo.hbmMask);
    end;
  end;
{$ENDIF}

type PHintInfo=^THintInfo;
var
  HintInfoPtr:PHintInfo;
{$IFDEF VER100}
    Spot:TSpot;
{$ENDIF}

begin

 HintInfoPtr:=PHintInfo(Message.LParam);

{$IFDEF VER100}
{  if FSpotIndex>=0 then}
   Spot:=FHotSpots[FSpotIndex]
{   else Spot:=nil}
;
  if assigned(FOnGetHintText)then
     FOnGetHintText(Self,Spot,HintInfoPtr^.HintStr);
{$ELSE}
  {aligning Hint position (bug/feature of VCL in Delphi 2.0)}
   With HintInfoPtr^ do
   begin
     GetCursorPos(HintPos);
     Inc(HintPos.Y,GetCursorHeightMargin);
   end;
{$ENDIF}
  Message.Result:=0;
end;






procedure THotImage.CMHitTest(var Message:TCMHitTest);

function ConvertPoint(X,Y:integer):TPoint;
var D,S,Offset:TPoint;
begin
 if (Picture.Graphic=nil) or (Picture.Graphic.Empty) then
 begin
  result.X:=-1;
  result.Y:=-1;
  exit;
 end;

 With DestRect do
 begin
  D.X:=Right-Left;
  D.Y:=Bottom-Top;
  Offset:=TopLeft;
 end;
 With Picture.Graphic do
 begin
  S.X:=Width;
  S.Y:=Height;
 end;
  Result.X:=MulDiv(X-Offset.X,S.X,D.X);
  Result.Y:=MulDiv(Y-Offset.Y,S.Y,D.Y);

end;


var i:integer;
   {$IFNDEF VER100}
    TempStr:string;
    Spot:TSpot;

   {$ENDIF}

begin

 With Message do
 begin
  i:=FHotSpots.IndexByPoint(ConvertPoint(XPos,YPos));
  if i<>FSpotIndex then
  begin
    Application.CancelHint;
    if i>=0 then
    begin
    {$IFNDEF VER100}
      Spot:=FHotSpots[i];
      TempStr:=Spot.FHintText;
      {the only place to get hint text with VCL from Delphi 2.0}
      if assigned(FOnGetHintText)then
       FOnGetHintText(Self,Spot,TempStr);
      inherited Hint:=TempStr;
    {$ELSE}
       inherited Hint:=FHotSpots[i].FHintText;
    {$ENDIF}
     end
    else
     inherited Hint:=FCommonHint;
   FSpotIndex:=i;

  end;
   result:=ord(i>=0);
 end;

end;

//hack method to get native TControlCanvas

type THackClass=class(TGraphicControl)
                 public
                  property Canvas;
               end;

function THotImage.GetCanvas:TCanvas;
begin
  result:=THackClass(Self).Canvas;
end;

function THotImage.GetPictureCanvas:TCanvas;
begin
 result:=inherited Canvas;
end;

procedure THotImage.DoPaint;
begin
 if assigned (FOnPaint) then FOnPaint(self);
end;

procedure THotImage.Paint;
begin
 inherited Paint;
 DoPaint;
end;

end.
