unit SpotEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,ExtCtrls,HotImage, Buttons;

type

  TfrmSpotEdit = class(TForm)
    GroupBox1: TGroupBox;
    ScrollBox1: TScrollBox;
    GroupBox2: TGroupBox;
    memHint: TMemo;
    Label1: TLabel;
    lboxPoints: TListBox;
    Label2: TLabel;
    btnOk: TButton;
    btnCancel: TButton;
    sbEdit: TSpeedButton;
    sbDeletePoint: TSpeedButton;
    Label3: TLabel;
    chkboxEnabled: TCheckBox;
    edSpotID: TEdit;
    rgZoom: TRadioGroup;
    Label4: TLabel;
    pnlShape: TPanel;
    sbPolyline: TSpeedButton;
    sbRect: TSpeedButton;
    sbElipse: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure sbEditClick(Sender: TObject);
    procedure ScrollBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure lboxPointsClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure sbDeletePointClick(Sender: TObject);
    procedure ScrollBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure edSpotIDKeyPress(Sender: TObject; var Key: Char);
    procedure rgZoomClick(Sender: TObject);
    procedure sbPolylineClick(Sender: TObject);
    procedure ScrollBox1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    { Private declarations }
    FGraphic:TGraphic;
    FSelectedKind:TSpotKind;
    FPaintRect:TRect;
    FNominator,
    FDenominator:integer;

    FPaintBox:TPaintBox;

    //points that  stores independed coordinates;
    FPoints:PPoints;
    FZmdPoints:PPoints;

    FPointsCount:integer;
    FSelectedPoint:Integer;


    FDrawing:Boolean;
    FOutOfControl:Boolean;
    FHiding:boolean;
    FOldPoint:TPoint;
    FOldSBProc:pointer;
    FNewSBProc:pointer ;
    procedure DoPaint(Sender:TObject);
    procedure FillLB;
    procedure SetSelectedPoint(Index:integer);
    procedure ResortPoints;
    procedure GetPoints(ASpot:TSpot);
    procedure PaintShape;
    procedure DeletePoint(index:integer);
    procedure AddPoint(const APoint:TPoint);
    procedure DrawEditLines;
    procedure HandleSBMessages(var Message:TMessage);

    procedure SetScale(Nominator,Denominator:integer);

    //Scale routines;
    function ItoD(const P:TPoint):TPoint;
    function DtoI(const P:TPoint):TPoint;
    procedure SetSpotKind(SpotKind:TSpotKind);
    procedure ClearAllPoints;

  public
    { Public declarations }
    function Edit(Graphic:TGraphic;Spot:TSpot):boolean;
  end;

var
  frmSpotEdit: TfrmSpotEdit;

implementation

{$R *.DFM}
{$R-}

type
  {Internal PaintBox without Mouse event handles}
  TPaintBoxInt=class(TPaintBox)
  protected
    procedure CMHitTest(var Message: TCMHitTest); message CM_HITTEST;
  end;



procedure TPaintBoxInt.CMHitTest(var Message: TCMHitTest);
begin
 Message.Result:=0;
end;

procedure DrawInvRect(const P:TPoint;Canvas:TCanvas);
begin
 With Canvas do
 begin
  Brush.Style:=bsSolid;
  Brush.Color:=clBlack;
 // Pen.Color:=clBlack;
  Pen.Mode:=pmNotXor;
  //Pen.Style:=psSolid;
  With p do
  Rectangle(X-4,Y-4,X+4,Y+4);
 end;
end;

procedure TfrmSpotEdit.HandleSBMessages(var Message:TMessage);
begin
With Message do
begin
 if Msg=WM_CAPTURECHANGED then
 begin
  label4.Caption:='';
   if FDrawing and not (FOutOfControl)then
   begin
    DrawEditLines;
    FHiding:=true;
   end;
 end
 else
  result:=CallWindowProc(FOldSBProc,ScrollBox1.Handle,Msg,WParam,LParam);
end;
end;

procedure TfrmSpotEdit.ResortPoints;
var buff:PPoints;
    nSize1,nSize2:Cardinal;

begin
 if (FSelectedPoint=-1) and (FPointsCount>0) then
   FSelectedPoint:=FPointsCount-1;
 if (FSelectedPoint<>FPointsCount-1) then
 begin
   nSize1:=sizeof(TPoint)*(FSelectedPoint+1);
   nSize2:=sizeof(TPoint)*FPointsCount-nSize1;

   Buff:=AllocMem(nSize1);
   Move(FPoints^,Buff^,nSize1);
   Move(FPoints[FSelectedPoint+1],FPoints^,nSize2);
   Move(buff^,FPoints[FPointsCount-FSelectedPoint-1],nSize1);
   FreeMem(buff);

   Buff:=AllocMem(nSize1);
   Move(FZmdPoints^,Buff^,nSize1);
   Move(FZmdPoints[FSelectedPoint+1],FZmdPoints^,nSize2);
   Move(buff^,FZmdPoints[FPointsCount-FSelectedPoint-1],nSize1);
   FreeMem(buff);

 end;

end;

procedure TfrmSpotEdit.DeletePoint(index:integer);
var
 nSize:integer;
begin
  PaintShape;
  lboxPoints.Items.Delete(index);
  nSize:=sizeof(TPoint)*(FPointsCount-1);
  if nSize>0 then
  begin
   If index<FPointsCount-1 then
   begin
    Move(FPoints[index+1],FPoints[index],nSize-sizeof(TPoint)*index);
    Move(FZmdPoints[index+1],FZmdPoints[index],nSize-sizeof(TPoint)*index);
   end;
   ReallocMem(FPoints,nSize);
   ReallocMem(FZmdPoints,nSize);
  end;
   Dec(FPointsCount);
   PaintShape;
end;

procedure TfrmSpotEdit.AddPoint(const APoint:TPoint);
var
 nSize:Integer;
begin
  PaintShape;

  nSize:=sizeof(TPoint)*(FPointsCount+1);
  if nSize>sizeof(TPoint) then
  begin
   ReallocMem(FPoints,nSize);
   ReallocMem(FZmdPoints,nSize);
  end;

  FPoints[FPointsCount]:=APoint;
  FZmdPoints[FPointsCount]:=ItoD(APoint);
  inc(FPointsCount);

  PaintShape;


//  FOldPoint:=APoint;
end;

procedure TfrmSpotEdit.SetSelectedPoint(Index:integer);
begin
  if not FDrawing then
  begin
   if FSelectedPoint>=0 then DrawInvRect(FZmdPoints[FSelectedPoint],FPaintBox.Canvas);
   if Index>=0 then DrawInvRect(FZmdPoints[Index],FPaintBox.Canvas);
  end;
  FSelectedPoint:=Index;
end;



procedure TfrmSpotEdit.DrawEditLines;
var TempPoint:TPoint;
    P:PPoint;
begin




 with FPaintBox.Canvas do
 begin
 TempPoint:=ItoD(FOldPoint);



 // Brush.Color:=clBlack;
  Brush.Style:=bsClear;
  Pen.Width:=1;
  Pen.Color:=clBlack;
  Pen.Mode:=pmNotXor;

  if FSelectedKind<>skPolyline then
  begin
   Pen.Style:=psSolid;
   MoveTo(0,TempPoint.Y);
   LineTo(FPaintBox.Width,TempPoint.Y);
   MoveTo(TempPoint.X,0);
   LineTo(TempPoint.X,FPaintBox.Height);
  end;

  if FPointsCount=0 then exit;

  Pen.Style:=psDot;
 if FSelectedKind= skPolyLine then
  begin
   with FZmdPoints[FPointsCount-1] do
    MoveTo(X,Y);
   With TempPoint do
    LineTo(X,Y);
   if FPointsCount>1 then
    with FZmdPoints[0]do
     LineTo(X,Y);
  end
  else
  begin
    if FSelectedPoint=0 then P:=@FZmdPoints[1]
    else P:=@FZmdPoints[0];
    case FSelectedKind of
     skRectangle:
      Rectangle(P^.X,P^.Y,TempPoint.X,TempPoint.Y);
     skElipse:
      Ellipse(P^.X,P^.Y,TempPoint.X,TempPoint.Y);
    end;
 end;
 end;//With Canvas
end;

procedure TfrmSpotEdit.PaintShape;
begin
if FPointsCount<2 then exit;{Nothing to draw}
With FPaintBox.Canvas do
begin
   Brush.Style:=bsClear;
   Pen.Color:=clBlack;
   Pen.Mode:=pmNotXor;
   Pen.Width:=2;
   Pen.Style:=psSolid;
   case FSelectedKind of
   skPolyline:
   begin
    Windows.Polyline(Handle,FZmdPoints^,FPointsCount);
    if (not FDrawing) and (FPointsCount>2) then
    begin
     begin

      With FZmdPoints[FPointsCount-1]do
       MoveTo(X,Y);
      With FZmdPoints[0]do
       LineTo(X,Y);
    end;
    end;
   end;
   skRectangle:
    if not FDrawing then
     Rectangle(FZmdPoints[0].X,FZmdPoints[0].Y,FZmdPoints[1].X,FZmdPoints[1].Y);

   skElipse:
     if not FDrawing then
       Ellipse(FZmdPoints[0].X,FZmdPoints[0].Y,FZmdPoints[1].X,FZmdPoints[1].Y);
   end;
end;
end;

procedure TfrmSpotEdit.DoPaint(Sender:TObject);
begin
 with FPaintBox.Canvas do
   StretchDraw(FPaintRect,FGraphic);
  PaintShape;
 if (FSelectedPoint>=0)and (not FDrawing) then
  DrawInvRect(FZmdPoints[FSelectedPoint],FPaintBox.Canvas);
end;

function PointToString(const P:TPoint):string;
begin
 result:=format('(%d,%d)',[P.X,P.Y]);
end;


procedure TfrmSpotEdit.FillLB;
var i:integer;
begin
 With lboxPoints.Items do
 begin
  BeginUpdate;
  try
   Clear;
   For i:=0 to FPointsCount-1 do
    Add(PointToString(FPoints[i]));
  finally
   EndUpdate;
  end;
 end;
 with lboxPoints do
  ItemIndex:=Items.Count-1 ;
 lboxPointsClick(self);
end;


procedure TfrmSpotEdit.GetPoints(ASpot:TSpot);
var
 nSize:integer;
 PR:PRect;
begin
 nSize:=ASpot.Count*sizeof(TPoint);
 if nSize>Sizeof(TPoint)  then
 begin
  ReAllocMem(FPoints,nSize);
  ReallocMem(FZmdPoints,nSize)
 end;
 if (ASpot.SpotKind<>skPolyline) then
 begin
  PR:=PRect(ASpot.GetPointArray);
  if (PR^.Left=PR^.Right) and (PR^.Top=PR^.Bottom) then
   FPointsCount:=0
  else
  begin
   FPointsCount:=2;
   Move(PR^,FPoints^,sizeof(TRect));
  end;
 end
 else
  begin
  if nSize>0 then
    Move(ASpot.GetPointArray^,FPoints^,ASpot.Count*sizeof(TPoint));
    FPointsCount:=ASpot.Count;
  end;
   //Still need to to Recalc FZmdPoints
end;


// Scale conversion for coordinates
function TfrmSpotEdit.ItoD(const P:TPoint):TPoint;
begin
 with p do
 begin
  result.X:=MulDiv(X,FNominator,FDenominator);
  result.Y:=MulDiv(Y,FNominator,FDenominator);
 end;
end;

function TfrmSpotEdit.DtoI(const P:TPoint):TPoint;
begin
 with p do
 begin
  result.X:=MulDiv(X,FDenominator,FNominator);
  result.Y:=MulDiv(Y,FDenominator,FNominator);
 end;
end;


procedure TfrmSpotEdit.SetScale(Nominator,Denominator:integer);

 function Max(X,Y:integer):integer;
 begin
  result:=X;
  if X<Y then result:=Y
 end;

var
 i,j:integer;

begin
 FNominator:=Nominator;
 FDenominator:=Denominator;
// Cancelling drawing if active
 ReleaseCapture;

 FPaintBox.Parent:=nil;
 //Setting painting rectangle for graphic
 FPaintRect:=Rect(0,0,MulDiv(FGraphic.Width,Nominator,Denominator),
                      MulDiv(FGraphic.Height,Nominator,Denominator));
 With FPaintRect,ScrollBox1 do

  FPaintBox.SetBounds(0,0,Max(Right,ClientWidth),Max(Bottom,ClientHeight));


 //Rescaling points for current Zoom-in ratio

 if FSelectedKind=skPolyline
 then j:=FPointsCount-1
 else j:=1;

 For i:=0 to j do
  With FPoints[i] do
  begin
   FZmdPoints[i].X:=MulDiv(X,Nominator,Denominator);
   FZmdPoints[i].Y:=MulDiv(Y,Nominator,Denominator);
  end;


 FPaintBox.Parent:=ScrollBox1;


 UpdateWindow(ScrollBox1.Handle);
end;

function TfrmSpotEdit.Edit(Graphic:TGraphic;Spot:TSpot):boolean;
var i:integer;
begin

 FGraphic:=Graphic;
 SetSpotKind(Spot.SpotKind);
 GetPoints(Spot);
 SetScale(1,1);

 chkboxEnabled.Checked:=Spot.Enabled;
 edSpotid.Text:=IntToStr(Spot.SpotID);
 memHint.Text:=Spot.HintText;

 FillLB;
 result:= ShowModal=mrOk;

 if result then
  with Spot do
  begin
   HintText:=MemHint.Text;
   Enabled:=chkboxEnabled.Checked;
   SpotID:=StrToIntDef(edSpotID.Text,0);
   Clear;
   if (FSelectedKind<>skPolyLine) and (FPointsCount<>2)
     then FPointsCount:=0;

   for i:=0 to FPointsCount-1 do
   AddPoint(FPoints[i]);

   SpotKind:=FSelectedKind;
  end;
end;

procedure TfrmSpotEdit.sbEditClick(Sender: TObject);
var SelPoint:integer;
begin
 if sbEdit.Down then
 begin
  ScrollBox1.Cursor:=crCross;
  lboxPoints.Enabled:=false;
  //BeginDrawing
  FOutOfControl:=true;
  PaintShape;
  SelPoint:=FSelectedPoint;
  SetSelectedPoint(-1);
  FSelectedPoint:=SelPoint;

  pnlShape.Enabled:=false;

  FDrawing:=true;
  FHiding:=true;

  if FSelectedKind=skPolyline then
  begin
   ResortPoints;
   FSelectedPoint:=-1;
   PaintShape;
  end;
 end
   else
   begin
   if (FSelectedKind<>skPolyLine) and (FPointsCount<>2) then FPointsCount:=0;
    pnlShape.Enabled:=true;
    PaintShape;
    FDrawing:=false;
    FSelectedPoint:=-1;
    PaintShape;
    ScrollBox1.Cursor:=crDefault;
    with lboxPoints do
     Enabled:=true;
    FillLB;
   end;
end;

procedure TfrmSpotEdit.sbDeletePointClick(Sender: TObject);
begin
 SetSelectedPoint(-1);
 if FSelectedKind=skPolyline then
 begin
  DeletePoint(lboxPoints.ItemIndex);
  lboxPointsClick(self);
 end
  else
  begin
   PaintShape;
   ClearAllPoints;
  end;
end;

{Mouse Messages}

procedure TfrmSpotEdit.ScrollBox1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 tempPoint:TPoint;

begin
 If (Button=mbRight) and FDrawing then
 begin
  ScrollBox1.Cursor:=crDefault;
  ReleaseCapture;
  SbEdit.Down:=false;
   sbEditClick(self);
 end


end;

procedure TfrmSpotEdit.ScrollBox1MouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 tempPoint:TPoint;
begin
 if (Button=mbLeft) then
 begin
    if FDrawing and (not FOutOfControl) then
  begin
   tempPoint:=DtoI(FPaintBox.ScreenToClient(ScrollBox1.ClientToScreen(point(X,Y))));
   if FSelectedKind=skPolyline then AddPoint(tempPoint)
   else
   begin
    if FPointsCount=0 then FSelectedPoint:=0;
    FPoints[FSelectedPoint]:=TempPoint;
    FZmdPoints[FSelectedPoint]:=ItoD(TempPoint);

    if  FPointsCount<2 then
    begin
     FSelectedPoint:=1-FPointsCount;
     inc(FPointsCount);
    end;
    if FPointsCount>=2 then
    begin
     ReleaseCapture;
     FSelectedPoint:=-1;

     ScrollBox1MouseDown(Sender,mbRight,Shift,X,Y);
    end;
   end;

 end;
end;
end;

procedure TfrmSpotEdit.ScrollBox1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
 tmpPoint:TPoint;
begin

 if FDrawing then
 begin
  if not ptInRect(ScrollBox1.ClientRect,Point(X,Y))then
  begin

    ReleaseCapture;
    FOutOfControl:=true;
  end
   else
    begin

     if(GetCapture<>ScrollBox1.Handle) then

      SetCapture(ScrollBox1.Handle);
      FOutOfControl:=false;


      if not FHiding then
        DrawEditLines;
      tmpPoint:=DtoI(FPaintBox.ScreenToClient(ScrollBox1.ClientToScreen(Point(X,Y))));
      Label4.Caption:=PointToString(tmpPoint);
      FOldPoint:=tmpPoint;
      DrawEditLines;
      FHiding:=false;

    end;
 end;
end;

procedure TfrmSpotEdit.lboxPointsClick(Sender: TObject);
var P:TPoint;
    R:TRect;
begin
 SetSelectedPoint(lboxPoints.ItemIndex);
  sbDeletePoint.Enabled:=(FSelectedPoint>=0);
//Show Active Selection
  if Sender=lboxPoints then
  begin
   p:=ScrollBox1.ScreenToClient(FPaintBox.ClientToScreen(FZmdPoints[FSelectedPoint]));
   R:=ScrollBox1.ClientRect;
   if not ptInRect(R,p) then
   begin
    if P.X>R.Right then P.X:=P.X-R.Right
    else if P.X>0 then P.X:=0;
    if P.Y>R.Bottom then P.Y:=P.Y-R.Bottom
    else if P.Y>0 then P.Y:=0;

    if P.X>0 then P.X:=P.X+R.Right div 2
    else
    if P.X<0 then P.X:=P.X-R.Right div 2 ;

    if P.Y>0 then P.Y:=P.Y+R.Bottom div 2
    else
     if P.Y<0 then P.Y:=P.Y-R.Bottom div 2 ;

     With ScrollBox1 do
     begin
      HorzScrollBar.Position:=HorzScrollBar.Position+P.x;
      VertScrollBar.Position:=VertScrollBar.Position+P.y;
     end;

   end;

  end;
end;

procedure TfrmSpotEdit.FormCreate(Sender: TObject);
begin
 FSelectedKind:=skPolyline;
 FSelectedPoint:=-1;
 FPaintBox:=TPaintBoxInt.Create(self);
 FPaintBox.OnPaint:=DoPaint;
 FPaintBox.Parent:=ScrollBox1;

 //Subclassing ScrollBox's Window
 FOldSBProc:=pointer(GetWindowLong(ScrollBox1.Handle,GWL_WNDPROC));
 FNewSBProc:=MakeObjectInstance(HandleSBMessages);
 SetWindowLong(ScrollBox1.Handle,GWL_WNDPROC,longint(FNewSBProc));

 //Allocating memory for point arrays
 FPoints:=AllocMem(sizeof(TPoint));
 FZmdPoints:=AllocMem(sizeof(TPoint));
end;

procedure TfrmSpotEdit.FormDestroy(Sender: TObject);
begin
 SetWindowLong(ScrollBox1.Handle,GWL_WNDPROC,longint(FOldSBProc));
 FreeObjectInstance(FNewSBProc);
 FreeMem(FPoints);
 FreeMem(FZmdPoints);

end;



procedure TfrmSpotEdit.edSpotIDKeyPress(Sender: TObject; var Key: Char);
begin
 if not (Key  in [#8,'0'..'9']) then
 Key:=#0;
end;

procedure TfrmSpotEdit.rgZoomClick(Sender: TObject);
var
 Denominator,Nominator:integer ;
begin
 case rgZoom.ItemIndex of
 0:{50%}
   begin
    Nominator:=1;
    Denominator:=2;
   end;
 1:{100%}
   begin
    Nominator:=1;
    Denominator:=1;
   end;
 2:{200%}
  begin
    Nominator:=2;
    Denominator:=1;
   end;
 3:{300%}
   begin
    Nominator:=3;
    Denominator:=1;
   end;
 else
    Nominator:=1;
    Denominator:=1;
 end;
 SetScale(Nominator,Denominator);
end;

procedure TfrmSpotEdit.ClearAllPoints;
begin
 SetSelectedPoint(-1);
 FPointsCount:=0;
 lboxPoints.Clear;
 sbDeletePoint.Enabled:=false;
end;


procedure TfrmSpotEdit.SetSpotkind(SpotKind:TSpotKind);
procedure RevertSpotkind;
var sb:TSpeedButton;
begin
 case FSelectedKind of
  skPolyline:sb:=sbPolyline;
  skRectangle:sb:=sbRect;
  else
   sb:=sbElipse;
  end;
 sb.Down:=true;
end;

var BS:cardinal;
begin
 if (SpotKind<>FSelectedKind) then
  If (FPointsCount=0) or
  (MessageDlg('Do you want to cancel current drawing?',
               mtWarning,[mbYes,mbNo],0)
  =mrYes)
  then
  begin
   PaintShape; //remove old drawing
   ClearAllPoints;
   FSelectedKind:=SpotKind;
   RevertSpotkind;//Sync button States
   if FSelectedKind=skPolyLine then
   begin
    sbDeletePoint.Caption:='De&lete point';
    ReallocMem(FPoints,sizeof(TPoint));
    ReallocMem(FZmdPoints,sizeof(TPoint));
   end
   else
    begin
     sbDeletePoint.Caption:='C&lear';
     ReallocMem(FPoints,sizeof(TPoint)*2);
     ReallocMem(FZmdPoints,sizeof(TPoint)*2);
    end;
  end
   else RevertSpotKind;

end;

procedure TfrmSpotEdit.sbPolylineClick(Sender: TObject);
var SK:TSpotKind;
begin
 if sbPolyline.Down then sk:=skPolyline else
  if sbRect.Down then sk:=skRectangle else
   sk:=skElipse;
  SetSpotKind(sk);
end;



end.
